#include <vector>
#include <iostream>
#include <assert.h>
#include <windows.h>
#include <shlwapi.h>
#include <wrl.h>
#include <dbghelp.h>

#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "dbghelp.lib")

using namespace Microsoft::WRL::Wrappers;

typedef HandleT<HandleTraits::HANDLENullTraits> ProcessHandle;

#if defined(_WIN64)

BOOL QueryTypeInformation(ProcessHandle& ProcessHandle, DWORD64 ModuleInstance, DWORD64 ExceptionTypeAddress, std::string& TypeInformationName, UINT32& TypeSize)
{
	DWORD ValuesA[4];
	SIZE_T ValuesASize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) ExceptionTypeAddress, ValuesA, sizeof ValuesA, &ValuesASize);
	assert(ValuesASize == sizeof ValuesA);

	DWORD64 CatchableTypesAddress = ModuleInstance + ValuesA[3]; // CatchableTypes RVA
	DWORD ValuesB[2];
	SIZE_T ValuesBSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) CatchableTypesAddress, ValuesB, sizeof ValuesB, &ValuesBSize);
	assert(ValuesBSize == sizeof ValuesB);

	assert(ValuesB[0] >= 1); // count
	DWORD64 CatchableTypeAddress = ModuleInstance + ValuesB[1]; // CatchableTypeAddress[0] RVA
	DWORD ValuesC[7];
	SIZE_T ValuesCSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) CatchableTypeAddress, ValuesC, sizeof ValuesC, &ValuesCSize);
	assert(ValuesCSize == sizeof ValuesC);

	DWORD64 TypeInformationAddress = ModuleInstance + ValuesC[1]; // std::type_info RVA
	BYTE DataD[1024];
	SIZE_T DataDSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) TypeInformationAddress, DataD, sizeof DataD, &DataDSize);
	assert(DataDSize == sizeof DataD);
	TypeInformationName = (const char*) (DataD + 0x10); // Decorated, see https://docs.microsoft.com/en-us/windows/desktop/api/dbghelp/nf-dbghelp-undecoratesymbolname

	//if(!TypeInformationName.empty())
	//{
	//	CHAR Name[512];
	//	if(UnDecorateSymbolName(TypeInformationName.c_str(), Name, _countof(Name), 0))
	//		;
	//}

	TypeSize = ValuesC[5];
	return TRUE;
}

#else

BOOL QueryTypeInformation(ProcessHandle& ProcessHandle, DWORD64 ModuleInstance, DWORD64 ExceptionTypeAddress, std::string& TypeInformationName, UINT32& TypeSize)
{
	DWORD ValuesA[4];
	SIZE_T ValuesASize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) ExceptionTypeAddress, ValuesA, sizeof ValuesA, &ValuesASize);
	assert(ValuesASize == sizeof ValuesA);

	DWORD64 CatchableTypesAddress = ValuesA[3]; // CatchableTypes RVA
	DWORD ValuesB[2];
	SIZE_T ValuesBSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) CatchableTypesAddress, ValuesB, sizeof ValuesB, &ValuesBSize);
	assert(ValuesBSize == sizeof ValuesB);

	assert(ValuesB[0] >= 1); // count
	DWORD64 CatchableTypeAddress = ValuesB[1]; // CatchableTypeAddress[0] RVA
	DWORD ValuesC[7];
	SIZE_T ValuesCSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) CatchableTypeAddress, ValuesC, sizeof ValuesC, &ValuesCSize);
	assert(ValuesCSize == sizeof ValuesC);

	DWORD64 TypeInformationAddress = ValuesC[1]; // std::type_info RVA
	BYTE DataD[1024];
	SIZE_T DataDSize = 0;
	ReadProcessMemory(ProcessHandle.Get(), (VOID*) TypeInformationAddress, DataD, sizeof DataD, &DataDSize);
	assert(DataDSize == sizeof DataD);
	TypeInformationName = (const char*) (DataD + 0x08); // Decorated, see https://docs.microsoft.com/en-us/windows/desktop/api/dbghelp/nf-dbghelp-undecoratesymbolname

	//if(!TypeInformationName.empty())
	//{
	//	CHAR Name[512];
	//	if(UnDecorateSymbolName(TypeInformationName.c_str(), Name, _countof(Name), 0))
	//		;
	//}

	TypeSize = ValuesC[5];
	return TRUE;
}

#endif

int main()
{
    //std::cout << "Hello World!\n"; 
	WCHAR Path[MAX_PATH];
	GetModuleFileName(nullptr, Path, _countof(Path));
	PathRemoveFileSpecW(Path);
	PathCombineW(Path, Path, L"Debuggee.exe");
	STARTUPINFOW StartupInfo = { sizeof StartupInfo };
	PROCESS_INFORMATION ProcessInformation = { };
	if(!CreateProcessW(Path, Path, nullptr, nullptr, TRUE, 0, nullptr, nullptr, &StartupInfo, &ProcessInformation))
		throw HRESULT_FROM_WIN32(GetLastError());
	CloseHandle(ProcessInformation.hThread);
	HandleT<HandleTraits::HANDLENullTraits> ProcessHandle(ProcessInformation.hProcess);
	if(!SymInitializeW(ProcessHandle.Get(), nullptr, FALSE)) // https://docs.microsoft.com/en-us/windows/desktop/api/dbghelp/nf-dbghelp-syminitialize
		throw HRESULT_FROM_WIN32(GetLastError());
	if(!DebugActiveProcess(ProcessInformation.dwProcessId))
		throw HRESULT_FROM_WIN32(GetLastError());
	for(BOOL Debug = TRUE; Debug; )
	{
		DEBUG_EVENT Event = { };
		if(!WaitForDebugEventEx(&Event, INFINITE))
			throw HRESULT_FROM_WIN32(GetLastError());
		WCHAR Text[4 << 10];
		swprintf_s(Text, L"Event.dwDebugEventCode %d\n", Event.dwDebugEventCode);
		OutputDebugStringW(Text);
		DWORD Status = DBG_CONTINUE;
		switch(Event.dwDebugEventCode)
		{
		#pragma region EXCEPTION_DEBUG_EVENT
		case EXCEPTION_DEBUG_EVENT:
			{
				Status = DBG_EXCEPTION_NOT_HANDLED;
		        EXCEPTION_RECORD& ExceptionRecord = Event.u.Exception.ExceptionRecord;
				swprintf_s(Text, L"dwFirstChance %d, ExceptionCode 0x%08X, ExceptionFlags 0x%X, ExceptionRecord 0x%p, ExceptionAddress 0x%p, NumberParameters %d", Event.u.Exception.dwFirstChance, ExceptionRecord.ExceptionCode, ExceptionRecord.ExceptionFlags, ExceptionRecord.ExceptionRecord, ExceptionRecord.ExceptionAddress, ExceptionRecord.NumberParameters);
				std::wstring TextEx = Text;
				for(DWORD Index = 0; Index < ExceptionRecord.NumberParameters; Index++)
				{
					swprintf_s(Text, L", 0x%p", (VOID*) ExceptionRecord.ExceptionInformation[Index]);
					TextEx.append(Text);
				}
				TextEx.append(L"\n");
				OutputDebugStringW(TextEx.c_str());
				// ExceptionFlags EXCEPTION_NONCONTINUABLE
				if(ExceptionRecord.ExceptionCode == 0x80000003)
					break;
				if(ExceptionRecord.ExceptionCode == 0xE06D7363) // C++ EH exception, https://blogs.msdn.microsoft.com/oldnewthing/20100730-00/?p=13273
				{
					const DWORD64 ObjectAddress = (DWORD64) ExceptionRecord.ExceptionInformation[1];
					const DWORD64 ObjectTypeAddress = (DWORD64) ExceptionRecord.ExceptionInformation[2];
					std::wstring ModuleImageName;
					#if defined(_WIN64)
						assert(ExceptionRecord.NumberParameters >= 4);
						const DWORD64 ModuleInstance = (DWORD64) ExceptionRecord.ExceptionInformation[3];
						SymRefreshModuleList(ProcessHandle.Get());
						IMAGEHLP_MODULEW64 ModuleInformation = { sizeof ModuleInformation };
						if(SymGetModuleInfoW64(ProcessHandle.Get(), ModuleInstance, &ModuleInformation))
						{
							ModuleImageName = ModuleInformation.ImageName;
						}
					#else
						assert(ExceptionRecord.NumberParameters >= 3);
						SymRefreshModuleList(ProcessHandle.Get());
						const DWORD64 ModuleInstance = SymGetModuleBase64(ProcessHandle.Get(), ObjectTypeAddress);
					#endif
					std::string SymbolName;
					DWORD64 Displacement;
					struct
					{
						IMAGEHLP_SYMBOL64 SymbolInformation;
						CHAR Name[256];
					} SymbolInformationAndName = { };
					SymbolInformationAndName.SymbolInformation.SizeOfStruct = sizeof SymbolInformationAndName.SymbolInformation;
					SymbolInformationAndName.SymbolInformation.MaxNameLength = (sizeof SymbolInformationAndName - offsetof(IMAGEHLP_SYMBOLW64, Name)) / sizeof SymbolInformationAndName.SymbolInformation.Name;
					if(SymGetSymFromAddr64(ProcessHandle.Get(), ObjectTypeAddress, &Displacement, &SymbolInformationAndName.SymbolInformation))
					{
						SymbolName = SymbolInformationAndName.SymbolInformation.Name;
						//CHAR Name[512];
						//if(UnDecorateSymbolName(SymbolName.c_str(), Name, _countof(Name), 0))
						//	;
					}
					std::string TypeInformationName;
					UINT32 TypeSize;
					#if defined(_WIN64)
						if(!QueryTypeInformation(ProcessHandle, ModuleInstance, ObjectTypeAddress, TypeInformationName, TypeSize))
							TypeSize = 0;
					#else
						if(!QueryTypeInformation(ProcessHandle, ModuleInstance, ObjectTypeAddress, TypeInformationName, TypeSize))
							TypeSize = 0;
					#endif
					std::wstring Value;
					if(TypeSize)
					{
						BYTE InstanceData[1024];
						SIZE_T InstanceDataSize = 0;
						if(ReadProcessMemory(ProcessHandle.Get(), (VOID*) ObjectAddress, InstanceData, min(TypeSize, sizeof InstanceData), &InstanceDataSize))
						{
							if(TypeInformationName.compare(".J") == 0) // LONG = HRESULT
							{
								if(InstanceDataSize == sizeof (LONG))
								{
									swprintf_s(Text, L"0x%08X", *((const LONG*) InstanceData));
									Value = Text;
								}
							}
							if(TypeInformationName.compare(".?AV_com_error@@") == 0) // LONG = HRESULT
							{
								if(InstanceDataSize >= sizeof (VOID*) + sizeof (LONG))
								{
									swprintf_s(Text, L"0x%08X", *((const LONG*) (InstanceData + sizeof (VOID*))));
									Value = Text;
								}
							}
						}
					}
					swprintf_s(Text, L"Module 0x%p %ls, SymbolName \"%hs\", TypeInformationName \"%hs\", %ls\n", (VOID*) ModuleInstance, ModuleImageName.c_str(), SymbolName.c_str(), TypeInformationName.c_str(), Value.c_str());
					OutputDebugStringW(Text);
				} else
				if(ExceptionRecord.ExceptionCode == 0x40080201) // RoOriginateError
				{
					assert(ExceptionRecord.NumberParameters >= 3);
					HRESULT Code = (HRESULT) ExceptionRecord.ExceptionInformation[0];
					UINT MessageLength = (UINT) ExceptionRecord.ExceptionInformation[1];
					DWORD64 MessageAddress = (DWORD64) ExceptionRecord.ExceptionInformation[2];
					BYTE MessageData[8 << 10];
					SIZE_T MessageDataSize;
					if(ReadProcessMemory(ProcessHandle.Get(), (VOID*) MessageAddress, MessageData, min(MessageLength * sizeof (WCHAR), sizeof MessageData), &MessageDataSize))
					{
						((WCHAR*) MessageData)[sizeof MessageData / sizeof (WCHAR) - 1] = 0;
						((WCHAR*) MessageData)[MessageLength] = 0;
						std::wstring Message = (WCHAR*) MessageData;
						Message.append(L"\n");
						OutputDebugStringW(Message.c_str());
					}
				}
			}
			break;
		#pragma endregion
		#pragma region EXIT_PROCESS_DEBUG_EVENT
		case EXIT_PROCESS_DEBUG_EVENT:
			{
				Debug = FALSE;
			}
			break;
		#pragma endregion
		#pragma region LOAD_DLL_DEBUG_EVENT
		case LOAD_DLL_DEBUG_EVENT:
			{
		        LOAD_DLL_DEBUG_INFO& LoadDll = Event.u.LoadDll;
				std::wstring DllImageName;
				if(LoadDll.lpImageName)
				{
					UINT64 ImageNamePointer;
					SIZE_T ImageNamePointerSize = 0;
					if(ReadProcessMemory(ProcessHandle.Get(), LoadDll.lpImageName, &ImageNamePointer, sizeof ImageNamePointer, &ImageNamePointerSize))
						if(ImageNamePointerSize == sizeof ImageNamePointer && ImageNamePointer)
						{
							WCHAR ImageName[MAX_PATH];
							SIZE_T ImageNameSize = 0;
							if(ReadProcessMemory(ProcessHandle.Get(), (VOID*) ImageNamePointer, ImageName, sizeof ImageName, &ImageNameSize))
								if(ImageNameSize)
								{
									assert(ImageNameSize <= sizeof ImageName);
									assert(LoadDll.fUnicode);
									DllImageName = ImageName;
								}
						}
				}
				swprintf_s(Text, L"hFile 0x%p, lpBaseOfDll 0x%p, dwDebugInfoFileOffset 0x%X, nDebugInfoSize 0x%X, lpImageName 0x%p %ls, fUnicode %d\n", LoadDll.hFile, LoadDll.lpBaseOfDll, LoadDll.dwDebugInfoFileOffset, LoadDll.nDebugInfoSize, LoadDll.lpImageName, DllImageName.c_str(), LoadDll.fUnicode);
				OutputDebugStringW(Text);
			}
			break;
		#pragma endregion
		#pragma region OUTPUT_DEBUG_STRING_EVENT
		case OUTPUT_DEBUG_STRING_EVENT:
			{
				if(Event.u.DebugString.fUnicode)
				{
					WCHAR Text[4 << 10];
					SIZE_T TextSize = 0;
					if(ReadProcessMemory(ProcessHandle.Get(), Event.u.DebugString.lpDebugStringData, Text, min(_countof(Text) - 1, (SIZE_T) Event.u.DebugString.nDebugStringLength) * sizeof *Text, &TextSize))
						if(TextSize)
						{
							assert(TextSize / sizeof *Text <= _countof(Text) - 1);
							Text[TextSize / sizeof *Text] = 0;
							std::wstring TextEx = Text;
							if(!TextEx.empty())
							{
								if(TextEx[TextEx.size() - 1] != L'\n')
									TextEx.append(L"\n");
								OutputDebugStringW(TextEx.c_str());
							}
						}
				}
			}
			break;
		#pragma endregion
		}
		if(!ContinueDebugEvent(Event.dwProcessId, Event.dwThreadId, Status))
			throw HRESULT_FROM_WIN32(GetLastError());
	}
	DebugActiveProcessStop(ProcessInformation.dwProcessId);
}
